/*
 * usb.c  -- USB procedures
 *
 * $Id: usb.c,v 1.7 2003/09/21 17:37:36 hos Exp $
 *
 */


#include "main.h"


NTSTATUS usb_reset(IN PDEVICE_OBJECT dev_obj);
NTSTATUS usb_get_dev_desc(IN PDEVICE_OBJECT dev_obj,
                          OUT PUSB_DEVICE_DESCRIPTOR *dev_desc,
                          OUT ULONG *size);
NTSTATUS usb_select_conf(IN PDEVICE_OBJECT dev_obj);
VOID DDKAPI usb_worker(IN PDEVICE_OBJECT dev_obj, IN PVOID ctxt);

NTSTATUS usb_call(IN PDEVICE_OBJECT dev_obj, IN ULONG ioctl_code,
                  IN PVOID arg1, IN PVOID arg2);


NTSTATUS device_init(IN PDEVICE_OBJECT dev_obj, IN PIRP irp)
{
    NTSTATUS nt_ret;
    device_extension_t *dev_ext;
    PUSB_DEVICE_DESCRIPTOR dev_desc;
    ULONG desc_size;

    dev_ext = GET_DEV_EXT(dev_obj);

    nt_ret = usb_reset(dev_obj);
    if(! NT_SUCCESS(nt_ret)) {
        return nt_ret;
    }

    nt_ret = usb_get_dev_desc(dev_obj, &dev_desc, &desc_size);
    if(! NT_SUCCESS(nt_ret)) {
        return nt_ret;
    }

    dev_ext->product_id = dev_desc->idProduct;
    dev_ext->vendor_id = dev_desc->idVendor;
    ExFreePool(dev_desc);

    nt_ret = usb_select_conf(dev_obj);
    if(! NT_SUCCESS(nt_ret)) {
        return nt_ret;
    }

    nt_ret = hid_init(dev_obj);
    if(! NT_SUCCESS(nt_ret)) {
        return nt_ret;
    }

    {
        PIO_WORKITEM work;

        work = IoAllocateWorkItem(dev_obj);
        if(work == NULL) {
            return STATUS_UNSUCCESSFUL;
        }

        IoQueueWorkItem(work, usb_worker, DelayedWorkQueue, work);
    }

    return STATUS_SUCCESS;
}

NTSTATUS device_remove(IN PDEVICE_OBJECT dev_obj)
{
    device_extension_t *dev_ext;

    dev_ext = GET_DEV_EXT(dev_obj);

    dev_ext->surprise_removedp = TRUE;

    return STATUS_SUCCESS;
}


NTSTATUS usb_get_port_status(IN PDEVICE_OBJECT dev_obj,
                             OUT ULONG *port_status);
NTSTATUS usb_reset_port(IN PDEVICE_OBJECT dev_obj);

NTSTATUS usb_reset(IN PDEVICE_OBJECT dev_obj)
{
    NTSTATUS nt_ret;
    ULONG port_status;

    nt_ret = usb_get_port_status(dev_obj, &port_status);
    if(! NT_SUCCESS(nt_ret)) {
        return nt_ret;
    }

    /* not connected */
    if((port_status & USBD_PORT_CONNECTED) == 0) {
        return STATUS_NO_SUCH_DEVICE;
    }

    /* already enabled */
    if((port_status & USBD_PORT_ENABLED) != 0) {
        return STATUS_SUCCESS;
    }

    /* reset port */
    nt_ret = usb_reset_port(dev_obj);
    if(! NT_SUCCESS(nt_ret)) {
        return nt_ret;
    }

    nt_ret = usb_get_port_status(dev_obj, &port_status);
    if(! NT_SUCCESS(nt_ret)) {
        return nt_ret;
    }

    if((port_status & USBD_PORT_CONNECTED) == 0 ||
       (port_status & USBD_PORT_ENABLED) == 0) {
        return STATUS_NO_SUCH_DEVICE;
    }

    return STATUS_SUCCESS;
}

NTSTATUS usb_get_port_status(IN PDEVICE_OBJECT dev_obj,
                             OUT ULONG *port_status)
{
    NTSTATUS nt_ret;

    nt_ret = usb_call(dev_obj, IOCTL_INTERNAL_USB_GET_PORT_STATUS,
                      port_status, NULL);

    return nt_ret;
}

NTSTATUS usb_reset_port(IN PDEVICE_OBJECT dev_obj)
{
    NTSTATUS nt_ret;

    nt_ret = usb_call(dev_obj, IOCTL_INTERNAL_USB_RESET_PORT,
                      NULL, NULL);

    return nt_ret;
}


NTSTATUS usb_get_dev_desc(IN PDEVICE_OBJECT dev_obj,
                          OUT PUSB_DEVICE_DESCRIPTOR *dev_desc,
                          OUT ULONG *size)
{
    NTSTATUS nt_ret;
    SIZE_T urb_size, desc_size;
    PURB urb;
    PUSB_DEVICE_DESCRIPTOR desc;

    urb_size = sizeof(struct _URB_CONTROL_DESCRIPTOR_REQUEST);
    urb = (PURB)ExAllocatePool(NonPagedPool, urb_size);
    if(urb == NULL) {
        return STATUS_INSUFFICIENT_RESOURCES;
    }

    desc_size = sizeof(USB_DEVICE_DESCRIPTOR);
    desc = (PUSB_DEVICE_DESCRIPTOR)ExAllocatePool(NonPagedPool, desc_size);
    if(desc == NULL) {
        ExFreePool(urb);
        return STATUS_INSUFFICIENT_RESOURCES;
    }

    UsbBuildGetDescriptorRequest(urb, urb_size,
                                 USB_DEVICE_DESCRIPTOR_TYPE, 0, 0,
                                 desc, NULL, desc_size,
                                 NULL);

    nt_ret = usb_call(dev_obj, IOCTL_INTERNAL_USB_SUBMIT_URB, urb, NULL);
    if(! NT_SUCCESS(nt_ret) || ! USBD_SUCCESS(urb->UrbHeader.Status)) {
        ExFreePool(urb);
        return STATUS_UNSUCCESSFUL;
    }

    *size = urb->UrbControlDescriptorRequest.TransferBufferLength;
    *dev_desc = desc;

    ExFreePool(urb);

    return nt_ret;
}


NTSTATUS usb_get_conf_desc(IN PDEVICE_OBJECT dev_obj,
                           OUT PUSB_CONFIGURATION_DESCRIPTOR *conf_desc,
                           OUT ULONG *size,
                           IN UCHAR conf_idx);

NTSTATUS usb_select_conf(IN PDEVICE_OBJECT dev_obj)
{
    device_extension_t *dev_ext;
    PUSB_CONFIGURATION_DESCRIPTOR conf_desc;
    PUSB_INTERFACE_DESCRIPTOR id;
    USBD_INTERFACE_LIST_ENTRY iface_list[2];
    PURB urb;
    PUSBD_INTERFACE_INFORMATION iface_info;
    ULONG size;
    NTSTATUS nt_ret;

    dev_ext = GET_DEV_EXT(dev_obj);

    nt_ret = usb_get_conf_desc(dev_obj, &conf_desc, &size, 0);
    if(! NT_SUCCESS(nt_ret)) {
        return nt_ret;
    }

    id = USBD_ParseConfigurationDescriptorEx(conf_desc, conf_desc,
                                             -1, -1,
                                             -1, -1, -1);
    if(id == NULL) {
        ExFreePool(conf_desc);
        return STATUS_NO_SUCH_DEVICE;
    }

    iface_list[0].InterfaceDescriptor = id;
    iface_list[0].Interface =NULL;
    iface_list[1].InterfaceDescriptor = NULL;

    urb = USBD_CreateConfigurationRequestEx(conf_desc, iface_list);

    nt_ret = usb_call(dev_obj, IOCTL_INTERNAL_USB_SUBMIT_URB, urb, NULL);
    if(! NT_SUCCESS(nt_ret) || ! USBD_SUCCESS(urb->UrbHeader.Status)) {
        ExFreePool(urb);
        ExFreePool(conf_desc);
        return STATUS_UNSUCCESSFUL;
    }

    dev_ext->usb_conf_handle = urb->UrbSelectConfiguration.ConfigurationHandle;
    iface_info = &urb->UrbSelectConfiguration.Interface;

    dev_ext->usb_input_pipe = NULL;
    dev_ext->usb_output_pipe = NULL;
    if(iface_info->NumberOfPipes >= 2) {
        PUSBD_PIPE_INFORMATION pipe_info;

        pipe_info = &iface_info->Pipes[0];
        dev_ext->usb_input_pipe = pipe_info->PipeHandle;

        pipe_info = &iface_info->Pipes[1];
        dev_ext->usb_output_pipe = pipe_info->PipeHandle;
    }

    ExFreePool(urb);
    ExFreePool(conf_desc);

    if(dev_ext->usb_input_pipe == NULL || dev_ext->usb_output_pipe == NULL) {
        return STATUS_UNSUCCESSFUL;
    }

    return STATUS_SUCCESS;
}

NTSTATUS usb_get_conf_desc(IN PDEVICE_OBJECT dev_obj,
                           OUT PUSB_CONFIGURATION_DESCRIPTOR *conf_desc,
                           OUT ULONG *size,
                           IN UCHAR conf_idx)
{
    NTSTATUS nt_ret;
    SIZE_T urb_size, desc_size;
    PURB urb;
    PUSB_CONFIGURATION_DESCRIPTOR desc;

    urb_size = sizeof(struct _URB_CONTROL_DESCRIPTOR_REQUEST);
    urb = (PURB)ExAllocatePool(NonPagedPool, urb_size);
    if(urb == NULL) {
        return STATUS_INSUFFICIENT_RESOURCES;
    }

    desc_size = sizeof(USB_CONFIGURATION_DESCRIPTOR);
    desc = (PUSB_CONFIGURATION_DESCRIPTOR)ExAllocatePool(NonPagedPool,
                                                         desc_size);
    if(desc == NULL) {
        ExFreePool(urb);
        return STATUS_INSUFFICIENT_RESOURCES;
    }

    UsbBuildGetDescriptorRequest(urb, urb_size,
                                 USB_CONFIGURATION_DESCRIPTOR_TYPE,
                                 conf_idx, 0,
                                 desc, NULL, desc_size,
                                 NULL);

    nt_ret = usb_call(dev_obj, IOCTL_INTERNAL_USB_SUBMIT_URB, urb, NULL);
    if(! NT_SUCCESS(nt_ret) || ! USBD_SUCCESS(urb->UrbHeader.Status)) {
        ExFreePool(urb);
        ExFreePool(desc);
        return STATUS_UNSUCCESSFUL;
    }

    desc_size = desc->wTotalLength;
    ExFreePool(desc);
    desc = (PUSB_CONFIGURATION_DESCRIPTOR)ExAllocatePool(NonPagedPool,
                                                         desc_size + 16); /* TODO: what is 16 ? */
    if(desc == NULL) {
        ExFreePool(urb);
        return STATUS_INSUFFICIENT_RESOURCES;
    }

    UsbBuildGetDescriptorRequest(urb, urb_size,
                                 USB_CONFIGURATION_DESCRIPTOR_TYPE,
                                 conf_idx, 0,
                                 desc, NULL, desc_size,
                                 NULL);

    nt_ret = usb_call(dev_obj, IOCTL_INTERNAL_USB_SUBMIT_URB, urb, NULL);
    if(! NT_SUCCESS(nt_ret) || ! USBD_SUCCESS(urb->UrbHeader.Status)) {
        ExFreePool(urb);
        ExFreePool(desc);
        return STATUS_UNSUCCESSFUL;
    }

    *size = urb->UrbControlDescriptorRequest.TransferBufferLength;
    *conf_desc = desc;

    ExFreePool(urb);

    return STATUS_SUCCESS;
}


VOID DDKAPI usb_worker(IN PDEVICE_OBJECT dev_obj, IN PVOID ctxt)
{
    device_extension_t *dev_ext;
    PIO_WORKITEM work;
    BYTE buf[sizeof(struct controller_raw_input)];
    ULONG buf_size;

    dev_ext = GET_DEV_EXT(dev_obj);
    work = (PIO_WORKITEM)ctxt;

    if(dev_ext->removedp) {
        return;
    }

    buf_size = sizeof(buf);
    usb_get_interrupt_transfer(dev_obj, buf, &buf_size);

    hid_update_from_raw(dev_obj, buf, sizeof(buf));

    IoFreeWorkItem(work);

    work = IoAllocateWorkItem(dev_obj);
    if(work == NULL) {
        return;
    }

    IoQueueWorkItem(work, usb_worker, DelayedWorkQueue, work);
}


NTSTATUS usb_interrupt_transfer(IN PDEVICE_OBJECT dev_obj,
                               IN PVOID buf, ULONG *buf_size,
                               USBD_PIPE_HANDLE pipe, ULONG flags);

NTSTATUS usb_get_interrupt_transfer(IN PDEVICE_OBJECT dev_obj,
                                   IN PVOID buf, ULONG *buf_size)
{
    device_extension_t *dev_ext;

    dev_ext = GET_DEV_EXT(dev_obj);

    return usb_interrupt_transfer(dev_obj, buf, buf_size,
                                 dev_ext->usb_input_pipe,
                                 (USBD_TRANSFER_DIRECTION_IN |
                                  USBD_SHORT_TRANSFER_OK));
}

NTSTATUS usb_put_interrupt_transfer(IN PDEVICE_OBJECT dev_obj,
                                   IN PVOID buf, ULONG *buf_size)
{
    device_extension_t *dev_ext;

    dev_ext = GET_DEV_EXT(dev_obj);

    return usb_interrupt_transfer(dev_obj, buf, buf_size,
                                 dev_ext->usb_output_pipe,
                                 0);
}

NTSTATUS usb_interrupt_transfer(IN PDEVICE_OBJECT dev_obj,
                               IN PVOID buf, ULONG *buf_size,
                               USBD_PIPE_HANDLE pipe, ULONG flags)
{
    SIZE_T urb_size;
    PURB urb;
    NTSTATUS nt_ret;

    if(pipe == NULL) {
        return STATUS_INVALID_HANDLE;
    }

    if(buf == NULL || *buf_size < 1) {
        return STATUS_INVALID_PARAMETER;
    }

    urb_size = sizeof(struct _URB_BULK_OR_INTERRUPT_TRANSFER);
    urb = (PURB)ExAllocatePool(NonPagedPool, urb_size);
    if(urb == NULL) {
        return STATUS_INSUFFICIENT_RESOURCES;
    }

    UsbBuildInterruptOrBulkTransferRequest(urb, urb_size,
                                           pipe,
                                           buf, NULL, *buf_size,
                                           flags,
                                           NULL);
    nt_ret = usb_call(dev_obj, IOCTL_INTERNAL_USB_SUBMIT_URB, urb, NULL);
    if(! NT_SUCCESS(nt_ret) || ! USBD_SUCCESS(urb->UrbHeader.Status)) {
        ExFreePool(urb);
        return STATUS_UNSUCCESSFUL;
    }

    *buf_size = urb->UrbBulkOrInterruptTransfer.TransferBufferLength;

    ExFreePool(urb);

    return STATUS_SUCCESS;
}


NTSTATUS usb_call(IN PDEVICE_OBJECT dev_obj, IN ULONG ioctl_code,
                  IN PVOID arg1, IN PVOID arg2)
{
    IO_STATUS_BLOCK io_status;
    PIRP irp;
    PIO_STACK_LOCATION irp_sp;
    KEVENT evt;
    NTSTATUS nt_ret;

    KeInitializeEvent(&evt, NotificationEvent, FALSE);

    irp = IoBuildDeviceIoControlRequest(ioctl_code,
                                        GET_NEXT_DEVICE_OBJECT(dev_obj),
                                        NULL, 0,
                                        NULL, 0,
                                        TRUE,
                                        &evt, &io_status);

    irp_sp = IoGetNextIrpStackLocation(irp);
    irp_sp->Parameters.Others.Argument1 = arg1;
    irp_sp->Parameters.Others.Argument2 = arg2;

    nt_ret = IoCallDriver(GET_NEXT_DEVICE_OBJECT(dev_obj), irp);
    if(nt_ret == STATUS_PENDING) {
        KeWaitForSingleObject(&evt, Suspended, KernelMode, FALSE, NULL);
        nt_ret = io_status.Status;
    }

    return nt_ret;
}
